import os
import errno


from botocore.exceptions import ClientError


from chappie.client.aws import S3Client
from chappie.client.local import LocalClient
from chappie.utils.random_value import RandomValue

class FileManager():

    def __init__(self, settings=None):
        """
            - settings parameter contain a dict with some specifications for storage service
            - Client is selected by CHAPPIE_STORAGE_SERVICE value, defined in the settings file
        """
        self.settings = settings
        self.storage_service = os.environ.get('CHAPPIE_STORAGE_SERVICE', 'local')
        self.bucket = os.environ.get('CHAPPIE_STORAGE_BUCKET_NAME', 'main-storage-bucket-nqhnuwm4u8jakcxu')

        # FileManager class has support for aws s3 and the local system
        if self.storage_service  == "local":
            self.client = LocalClient()
        elif self.storage_service == "s3":
            self.client = S3Client()
            try:
                self.client.client.head_bucket(Bucket=self.bucket)
            except ClientError:
                self.client.create_bucket(self.bucket)
        else:
            self.client = None
            raise ValueError("Storage Service (%s) is not supported"%(self.storage_service))

        self.params_dict = {"Service": self.storage_service}
        if self.settings:
            self.client.bucket = self.settings.get('Bucket')
        else:
            self.client.bucket = self.bucket

        self.client.base_dir = os.environ.get('CHAPPIE_BASE_PROJECT_DIR', os.path.abspath(os.path.join(os.path.dirname(os.path.dirname(__file__)), '..')))

        if self.storage_service == "local":
            try:
                os.makedirs(os.path.join(self.client.base_dir, 'file_manager', self.bucket))
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise e

    def generate_file_path(self, company_id, branch_id, app_name):
        """
        Create a FileManager instance with proper Settings (dependent on the storage Service)
        Call generate_file_path() to cerate standard path to file.
        Input: company_id, branch_id, app_name. (app_name is used to organize data in the storage service)
        Return value: string
        """
        self.file_path = "%s/%s/%s"%(company_id, branch_id, app_name)
        return self.file_path

    def upload_file(self, file_data, path, filename):
        """
        Create a FileManager instance with proper Settings (dependent on the storage Service)
        Call generate_file_path() to cerate standard path to file - In rare cases, set custom path
        Call upload_file() - This calls appropriate Service client's upload file method.
        input: file_data as a bytes array. path, filename as strings
        return value: dict with parameters indicating storage location
        """
        if isinstance(file_data, bytes):
            return self.client.upload_file(file_data, path, filename, self.params_dict)
        else:
            raise ValueError("file_data should be a bytes array")

    def download_file(self, params_dict):
        """
        Create a FileManager instance with proper Settings (dependent on the storage Service)
        Call generate_file_path() to cerate standard path to file - In rare cases, set custom path
        Call download_file() - This calls appropriate Service client's download file method.
        Input: dict with parameters indicating storage location
        return value: filedata as a bytes array
        """
        return self.client.download_file(params_dict)

    def get_url(self, params_dict):
        """
        Create a FileManager instance with proper Settings (dependent on the storage Service)
        Input: dict with parameters indicating storage location
        return value: a temporary authenticated url to download file data.
        NOTE: do not store this url for future use. Call get_url whenever needed
        """
        return self.client.get_url(params_dict)

    def delete_file(self, params_dict):
        """
        Create a FileManager instance with proper Settings (dependent on the storage Service)
        Input: dict with parameters indicating storage location
        return value: Boolean indicating success or failure.
        """
        return self.client.delete_file(params_dict)

    def create_bucket(self, bucket_name):
        """
        Create a FileManager instance with proper Settings (dependent on the storage Service)
        Input: bucket name
        return value: path to new created bucket
        """
        return self.client.create_bucket(bucket_name)

    def put_object(self, key):
        """
        Create a Key object inside a Bucket
        Input: Key
        return value: path to new created object
        """
        return self.client.put_object(key)



import boto3, json, os, requests, uuid, sys
from cerberus import Validator
from urllib.request import urlopen
from urllib.parse import urlparse

from .local_client import LocalClient


# TODO for completeness and test needs, might have to implement cached queues in local stack too
class LocalstackSNSClient():

    SNS_MESSAGE_TYPE_SUB_NOTIFICATION = "SubscriptionConfirmation"
    SNS_MESSAGE_TYPE_NOTIFICATION = "Notification"
    SNS_MESSAGE_TYPE_UNSUB_NOTIFICATION = "UnsubscribeConfirmation"

    MIN_MESSAGE_SIZE_TO_BUFFER = 48000 # in Bytes. Anything bigger than a small json.

    def __init__(self):
        # NOTE Setting local endpoint_url
        if os.environ.get('DOCKER_RUNTIME') is None:
            self.sns_client = boto3.client('sns', endpoint_url="http://localhost:4575")
            self.sqs_client = boto3.client('sqs', endpoint_url="http://localhost:4576")
        else:
            self.sns_client = boto3.client('sns', endpoint_url="http://host.docker.internal:4575")
            self.sqs_client = boto3.client('sqs', endpoint_url="http://host.docker.internal:4576")

        self.app_name = ''
        self.topic = ''
        self.protocol = ''
        self.endpoint_url = ''
        self.message_filter = '__all__'

    def initialize(self):
        """
        create queues based on Filters
            - apply subscription and filter policies
            - store list of created queues and associated TokenQueues as safe_queue_url_list
        Remove abandoned queues
            - list all queues starting with 'AppName_'
            - For all existing queues not in 'safe_queue_url _list', delete the queues and their associated token queue
            NOTE: AWS specification says that there must be a 60 sec delay before creating a queue with the same name as a deleted Queue

        Update subscription of MicroSerice to the SNS Topics
        """
        safe_queue_url_list = []
        if type(self.message_filter) == list:
            for event in self.message_filter:
                queue_url, token_queue_url = self.initialize_queue(event)
                safe_queue_url_list.append(queue_url)
                safe_queue_url_list.append(token_queue_url)

        existing_queues = self.sqs_client.list_queues(QueueNamePrefix=self.get_queue_name_prefix())

        if existing_queues.get("QueueUrls") is not None:
            for queue_url in existing_queues["QueueUrls"]:
                if queue_url not in safe_queue_url_list:
                    queue_attributes = self.sqs_client.get_queue_attributes(
                            QueueUrl=queue_url,
                            AttributeNames=['All']
                    )
                    queue_arn = queue_attributes["Attributes"].get('QueueArn')

                    try:
                        # self.sns_client.delete_endpoint(
                        #     EndpointArn=queue_arn
                        # )
                        self.initialize_subscription(self.topic, 'sqs', queue_arn, '__none__')
                        self.sqs_client.delete_queue(
                            QueueUrl=queue_url
                        )
                    except Exception as e:
                        print(e)

        self.initialize_subscription(self.topic, self.protocol, self.endpoint_url, self.message_filter)

    def initialize_queue(self, event):
        """
        Create a queue for each event
            Queue will subscribe to the SNS topic and filter messages by event Name
        A Token Queue is always created while creating a queue.
            Some downstream process (in Alma?) should handle creating/deleting tokens and managing concurrency of processess handling the event

        Ensure to purge queue after initialize
            - NOTE that purge queue might make system unstable for up to 60 seconds (Aws docs: https://docs.aws.amazon.com/cli/latest/reference/sqs/purge-queue.html )
        """
        fifo = event.get('fifo', False)
        if fifo == False:
            queue = self.sqs_client.create_queue(
                    QueueName=self.get_queue_name(event),
                    Attributes={
                        "VisibilityTimeout": str(event.get('visibility_timeout', 30)),
                        "ReceiveMessageWaitTimeSeconds": "1",
                    }
                )
        else:
            queue = self.sqs_client.create_queue(
                    QueueName=self.get_queue_name(event),
                    Attributes={
                        "VisibilityTimeout": str(event.get('visibility_timeout', 30)),
                        "ReceiveMessageWaitTimeSeconds": "1",
                        "FifoQueue": "true",
                        "ContentBasedDeduplication": "true"
                    }
                )

        token_queue = self.sqs_client.create_queue(
                QueueName=self.get_token_queue_name(event),
                Attributes={
                    "FifoQueue": "true",
                    "ContentBasedDeduplication": "true",
                    "ReceiveMessageWaitTimeSeconds": "1",
                }
            )

        queue_url = queue.get('QueueUrl')
        token_queue_url = token_queue.get('QueueUrl')

        try:
            self.sqs_client.purge_queue(QueueUrl=queue_url)
        except Exception as e:
            print(e)

        try:
            self.sqs_client.purge_queue(QueueUrl=token_queue_url)
        except Exception as e:
            print(e)

        if queue_url is not None:  ## NOTE this may be needed for error handling.
            queue_attributes = self.sqs_client.get_queue_attributes(
                    QueueUrl=queue_url,
                    AttributeNames=['All']
            )
            queue_arn = queue_attributes["Attributes"].get('QueueArn')

            if fifo == False:
                self.initialize_subscription(self.topic, 'sqs', queue_arn, [event])

            return queue_url, token_queue_url

        else:
            print("Unable to initialize queue for event:", event)
            return None, None

    def initialize_subscription(self, topic, protocol, endpoint_url, message_filter):
        """
        Find subscription. Create one if not found.
        Update filter policy to listen to all events in the message_filter
        """
        if endpoint_url != '' and topic != '':
            try:
                response = self.sns_client.list_subscriptions_by_topic(TopicArn=topic)
                next_token = response.get("NextToken")

                while True:
                    for subscription in response["Subscriptions"]:
                        if subscription["Endpoint"] == endpoint_url:
                            subscription_arn = subscription["SubscriptionArn"]
                            self.update_filter_policy(subscription_arn, message_filter)

                            return subscription_arn
                    if next_token is not None:
                        response = self.sns_client.list_subscriptions_by_topic(TopicArn=topic,
                                                                            NextToken=next_token)
                        next_token = response.get("NextToken")
                    else:
                        break

                # NOTE this happens only if no existing subscriptions are found
                if message_filter != '__none__':
                    # Do not subscribe
                    subscription = self.sns_client.subscribe(
                        TopicArn=topic,
                        Protocol=protocol,
                        Endpoint=endpoint_url,
                        ReturnSubscriptionArn=True
                    )
                    subscription_arn = subscription["SubscriptionArn"]
                    self.update_filter_policy(subscription_arn, message_filter)

                    return subscription_arn
            except Exception as e:
                print(e)
                return None
        else:
            return None

    def update_filter_policy(self, subscription_arn, message_filter):
        """
        Implements event filter for the subscription
        Set raw attribute for subscription (reduce message size)
        """

        if len(subscription_arn.split(':')) < 7:
            raise Exception("Invalid subscription: " + subscription_arn)

        if message_filter == "__none__":
            # unsubscribe
            try:
                self.sns_client.unsubscribe(SubscriptionArn=subscription_arn)
            except Exception as e:
                print(e)
            return
        elif message_filter == "__all__":
            filter_policy = {}
        else:
            filter_policy = {
                "fliegen_workflow_event": [item["name"] for item in message_filter]
            }

        try:
            self.sns_client.set_subscription_attributes(
                SubscriptionArn=subscription_arn,
                AttributeName="FilterPolicy",
                AttributeValue=json.dumps(filter_policy))
            self.sns_client.set_subscription_attributes(
                SubscriptionArn=subscription_arn,
                AttributeName="RawMessageDelivery",
                AttributeValue="true"
            )
        except Exception as e:
            print(e)

    def add_process(self, event):
        """
        Check if number of active process is < concurrency limit
        if yes:
            - increment process counter.
            - return process token
        if No:
            - return None
        """
        token_queue_url_dict = self.sqs_client.get_queue_url(QueueName=self.get_token_queue_name(event))
        token_queue_url = token_queue_url_dict["QueueUrl"]

        token_queue_attributes = self.sqs_client.get_queue_attributes(
                                        QueueUrl=token_queue_url,
                                        AttributeNames=['All']
                                    )
        num_processes = int(token_queue_attributes["Attributes"].get('ApproximateNumberOfMessages')) \
                            + int(token_queue_attributes["Attributes"].get('ApproximateNumberOfMessagesNotVisible'))
         # This is a fifo Queue so the number should be accurate. However there may be some discrepancy due to async access

        if num_processes < event.get('concurrency_limit', 1):
            print("Approved addition of FliegenToken (%d) to start task for '%s'"%(num_processes, event.get('name')))
            return self.sqs_client.send_message(
                QueueUrl=token_queue_url,
                MessageBody=str(uuid.uuid4()),
                MessageGroupId='FliegenToken'
            )

        else:
            return None

    def remove_process(self, event):
        """
        delete token and decrement the active process count
        return None
        """
        token_queue_url_dict = self.sqs_client.get_queue_url(QueueName=self.get_token_queue_name(event))
        token_queue_url = token_queue_url_dict["QueueUrl"]

        token_queue_attributes = self.sqs_client.get_queue_attributes(
                                        QueueUrl=token_queue_url,
                                        AttributeNames=['All']
                                    )
        num_processes = int(token_queue_attributes["Attributes"].get('ApproximateNumberOfMessages')) \
                            + int(token_queue_attributes["Attributes"].get('ApproximateNumberOfMessagesNotVisible'))

        response = self.sqs_client.receive_message(
                QueueUrl=token_queue_url,
        )

        # NOTE handle situation when message No message is recieved
        if response.get("Messages") is not None:
            messages_list = response.get("Messages")
            message_handle = messages_list[0]["ReceiptHandle"]

            print("Destroying FliegenToken for '%s'. %d Active Processes."%(event.get('name'), num_processes))
            return self.sqs_client.delete_message(
                QueueUrl=token_queue_url,
                ReceiptHandle=message_handle
            )
        else:
            return None

    def publish_message(self, message, fliegen_workflow_event='default'):
        """
        Send small messages are sent to the SNS Topic.
        Large messages are stored in /tmp in local file system using File manager
            - FileManager returns a params dict with storage data
        The FileManager return value is packaged in the message payload.
        """
        payload = json.dumps(message.get("payload"))
        if sys.getsizeof(payload) > self.MIN_MESSAGE_SIZE_TO_BUFFER:
            file_manager = LocalClient()
            file_manager.folder = '/tmp/'

            payload = file_manager.upload_file(
                    file_data=payload.encode(),
                    path=fliegen_workflow_event + '/',
                    filename=str(uuid.uuid4()),
                    params_dict={
                        "Service": 'local'
                    })

            message["payload"] = payload

        message = json.dumps(message) # TODO check which works for localtack
        # message = {"default": json.dumps(message)}
        message_attributes = {
            'fliegen_workflow_event': {
                    'StringValue': fliegen_workflow_event,
                    'DataType': 'String'
            }
        }
        if len(self.topic.split(':')) < 6:
            raise Exception("Invalid topic: " + self.topic)

        try:
            return_value = self.sns_client.publish(
                            TopicArn=self.topic,
                            Message=json.dumps(message),
                            MessageStructure='json',
                            MessageAttributes=message_attributes)
            return return_value
        except Exception as e:
            print(e)
            return None

    def extract_message(self, request):
        """
        NOTE Extract message does not pull actual message if it is buffered in file
            - this is done for code optimization

        Return message data that is packaged in a wrapper (for http endpoint, it's request object)
        """
        try:
            notification_message = json.loads(request.body)

            if type(notification_message) == str:
                message = json.loads(notification_message)
            elif type(notification_message) == dict:
                message = notification_message
            else:
                print("Unable to interpret Message: ", notification_message)
                message = None

            return message
        except Exception as e:
            raise(e)

    def retrieve_message(self, event):
        """
        Look into message queue and pull out the next message to process.
        If payload is a FileManager Params dict
            - read actual message form file.
            - replace payload in message
        return Message and message Handle
        If no more messages Exist return None
            - NOTE important as this is the termination criteria for recursion
        """
        buffered_message_template = {
            "Service": {"type": "string", "required": True},  # TODO change to report_name
            "Filename": {"type": "string", "required": True}
        }
        valid_buffered_message = Validator(buffered_message_template, allow_unknown=False)

        queue_url_dict = self.sqs_client.get_queue_url(QueueName=self.get_queue_name(event))
        queue_url = queue_url_dict["QueueUrl"]

        response = self.sqs_client.receive_message(
                QueueUrl=queue_url,
        )

        if response.get("Messages") is not None: #TODO check if this is the right condition
            messages_list = response.get("Messages")

            body = json.loads(messages_list[0]["Body"])
            if type(body) == str:
                body = json.loads(body)
            if body.get("Message") is not None:
                message = json.loads(body.get("Message"))
            else:
                message = body

            payload = message.get('payload')
            if valid_buffered_message.validate(payload):
                file_manager = LocalClient()
                file_manager.folder = '/tmp/'

                payload = file_manager.download_file(payload)
                payload = json.loads(payload)

                message["payload"] = payload

            return {
                "message": message,
                "message_handle": messages_list[0]["ReceiptHandle"]
            }
        else:
            print("Processed all messages from queue for event '%s'!"%(event['name']), response)
            return None

    def delete_message(self, event, message_handle):
        """
        NOTE: Buffered data are not deleted. Just periodically delete files from /tmp
        Delete message from queue given the message_handle
        return None
        """
        queue_url_dict = self.sqs_client.get_queue_url(QueueName=self.get_queue_name(event))
        queue_url = queue_url_dict["QueueUrl"]

        return self.sqs_client.delete_message(
            QueueUrl=queue_url,
            ReceiptHandle=message_handle
        )

    def push_message(self, message, queue_url):
        """
        Small messages are pushed into the queue as-is.
        Large messages are stored in in /tmp folder using File manager
            - FileManager returns a params dict with storage data
        The FileManager return value is packaged in the message payload.
        """
        payload = json.dumps(message.get("payload"))
        if sys.getsizeof(payload) > self.MIN_MESSAGE_SIZE_TO_BUFFER:
            file_manager = LocalClient()
            file_manager.folder = '/tmp/'

            message = file_manager.upload_file(
                    file_data=payload.encode(),
                    path=queue_url.split('/')[-1] + '/',
                    filename=str(uuid.uuid4()),
                    params_dict={
                        "Service": 'local'
                    })

            message["payload"] = payload

        return self.sqs_client.send_message( # NOTE watch out for deduplication crash, whatever that is...
            QueueUrl=queue_url,
            MessageBody=json.dumps(message),
            # MessageBody=message,
            MessageGroupId='FliegenPushFifo'
        )

    def get_queue_name_prefix(self):
        """
        Standardise prefix (helps with searching for queues)
        """
        app_name = self.app_name
        topic_name = self.topic.split(':')[-1]

        return app_name + '_' + topic_name + '_'

    def get_queue_name(self, event):
        """
        ensure fifo queues have .fifo suffix
        """
        if event.get('fifo', False):
            return self.get_queue_name_prefix() + event['name'] + '.fifo'
        else:
            return self.get_queue_name_prefix() + event['name'] # + '.fifo'

    def get_token_queue_name(self, event):
        """
        Keep token queue name similar to the queue name
        """
        return self.get_queue_name_prefix() + event['name'] + '_' + 'tokens.fifo'

    def get_event_queue(self, event):
        """
        Get the queue endpoint for a specific event
        """
        queue_url_dict = self.sqs_client.get_queue_url(QueueName=self.get_queue_name(event))
        return queue_url_dict["QueueUrl"]

    def process_event(self, event, message):
        """
        If event is declared as a FIFO queue, the message is pushed into its FIFO queue
            Standard queue messages are auto pushed to SQS via subscription.
        """
        if event.get('fifo', False):
            queue_url = self.get_event_queue(event=event)
            self.push_message(message=message, queue_url=queue_url)




from chappie.client.aws import SNSClient
from chappie.client.local import LocalstackSNSClient


class MessageManager():

    def __init__(self, params):
        """ params is a dictionary with 'Service' and other relevant details needed for the service
        For example for aws-sns,
        params = {
            "Service": "aws-sns",
            "Topic": "topic_arn",
            "Endpoint_Url": message listener endpoint,
            "Message_Filter": one of '__all__' or '__none__' or ["list", "of", "strings"]
        }
        For localstack messaging platform,
        params = {
            "Service": "localstack-sns",
            "Topic": "topic_arn",
            "Endpoint_Url": message listener endpoint,
            "Message_Filter": one of '__all__' or '__none__' or ["list", "of", "strings"]
        }
        For rabbitMQ,
        params = {
            "Service": "rabbitmq",
            "NEED": "TO FIND THE PARAMS"
        }
        """
        self.params = params
        self.service = self.params.get("Service")

        if self.service == "localstack-sns":
            self.client = LocalstackSNSClient()
            self.client.app_name = self.params.get('App_Name', '')
            self.client.topic = self.params.get('Topic', '')
            self.client.protocol = self.params.get('Protocol', '')
            self.client.endpoint_url = self.params.get('Endpoint_Url', '')
            self.client.message_filter = self.params.get('Message_Filter', '')
        elif self.service == "aws-sns":
            self.client = SNSClient()
            self.client.app_name = self.params.get('App_Name', '')
            self.client.topic = self.params.get('Topic', '')
            self.client.protocol = self.params.get('Protocol', '')
            self.client.endpoint_url = self.params.get('Endpoint_Url', '')
            self.client.message_filter = self.params.get('Message_Filter', '')
        else:
            self.client = None
            raise ValueError("Messaging Service (%s) is not supported"%(self.service))

    def initialize(self):
        """
        Initialize The messaging platform.
        For AWS, create Queues, Subscriptions, S3 Buffer bucket, and Filter policies
        For LocalStack, create Queues, Subscriptions, Local File storage, and Filter policies
        """
        return self.client.initialize()

    def publish_message(self, *,  message, fliegen_workflow_event='default'):
        """
        Publishes a message on the platform.
        Message will be published with the fliegen_workflow_event
            - For SQS it's added to Attributes, subscribers filter for the appropriate filter policy
        """
        return self.client.publish_message(message, fliegen_workflow_event)

    # TODO change to _from_request
    def extract_message(self, *, wrapper):
        """
        This is used by the message listener (subscriber) to interpret the message recieved at the endpoint.
        - NOTE Buffered messages are not retrieved here (to minimize S3 calls)
        """
        return self.client.extract_message(wrapper)

    # def update_filter_policy(self):
    #     return self.client.update_filter_policy()

    def get_event_queue(self, *, event):
        """
        Returns the Queue endpoint given an event dictionary
        """
        # Cerberus for event format
        return self.client.get_event_queue(event)

    def add_process(self, *, event):
        """
        Check if number of active process is < concurrency limit
        if yes:
            - increment process counter.
            - return process token
        if No:
            - return None
        """
        return self.client.add_process(event)

    def remove_process(self, *, event):
        """
        delete token and decrement the active process count
        return None
        """
        return self.client.remove_process(event)

    def push_message(self, *, message, queue_url):
        """
        Push a message into a queue.
        This is used to ensure fifo queues are handled in sequence
        """
        return self.client.push_message(message, queue_url)

    def retrieve_message(self, *, event):
        """
        Look into message queue and pull out the next message to process.
        return Message and message ID
        If no more messages Exist return None
            - NOTE important as this is the termination criteria for recursion
        """
        return self.client.retrieve_message(event)

    def delete_message(self, *, event, message_handle):
        """
        delete message from queue
        return None
        """
        return self.client.delete_message(event, message_handle)

    def process_event(self, *, event, message):
        """
        Handle any special requirements to process event specific to the client used.
        - For SQS fifo queue, push message to queue
        """
        return self.client.process_event(event, message)



# -*- coding: utf-8 -*-
import os, unittest
from io import BytesIO
from chappie.client.aws import S3Client

class TestS3Client(unittest.TestCase):
    def setUp(self):
        """ """
        self.client = S3Client()
        self.client.bucket = "bucket-test-w37ryifgckcbjqiu4yrf78qy3874yr8cwfc"

        self.params_dict = {
            "Service": 's3'
        }
        file_obj = BytesIO(b'test')
        self.params_dict = self.client.upload_file(file_obj, 'chappie-test/tmp/', 'test.txt', self.params_dict)

    def test_file_upload(self):
        """Test upload of file data"""

        self.assertIsInstance(self.params_dict, dict)
        self.assertEqual(self.params_dict.get('Bucket'), self.client.bucket)
        self.assertEqual(self.params_dict.get('Key'), 'chappie-test/tmp/test.txt')

    def test_file_download(self):
        """Test download of file data"""

        filedata = self.client.download_file(self.params_dict)

        self.assertIsInstance(filedata, bytes)
        self.assertEqual(filedata.decode(), 'test')

    def test_get_url(self):
        """Test get file access url"""

        url = self.client.get_url(self.params_dict)
        self.assertGreater(len(url), 0)

    def test_delete_file(self):
        """Test file delete"""

        status = self.client.delete_file(self.params_dict)





import os


import boto3
from io import BytesIO
from datetime import datetime
from dateutil.relativedelta import relativedelta

class S3Client():

    def __init__(self):
        boto3.setup_default_session(profile_name=os.environ.get('CHAPPIE_AWS_PROFILE', 'default'))
        self.client = boto3.client('s3')
        self.bucket = None
        self.expires = None # make sure this is None or number of hours

    def upload_file(self, file_data, path, filename, params_dict):
        """ Save Bytes array data into a bucket and return FileManager Params."""
        if self.expires:
            self.client.put_object(
                            Body=file_data,
                            Bucket=self.bucket,
                            Key='%s/%s' % (path, filename),
                            Expires=datetime.now() + relativedelta(hours=self.expires),
                        )
        else:
            self.client.put_object(
                            Body=file_data,
                            Bucket=self.bucket,
                            Key='%s/%s' % (path, filename),
                        )

        waiter = self.client.get_waiter('object_exists')
        try:
            waiter.wait(
                    Bucket=self.bucket,
                    Key='%s/%s' % (path, filename),
                )
        except Exception as e:
            print(e)
            return None

        return_params_dict = params_dict.copy()
        return_params_dict['Bucket'] = self.bucket
        return_params_dict['Key'] = '%s/%s' % (path, filename)

        return return_params_dict

    def download_file(self, params_dict):
        """ Return file data as a Bytes array"""
        file_data = BytesIO()
        self.client.download_fileobj(params_dict.get('Bucket'),
                                    params_dict.get('Key'),
                                    file_data)
        return file_data.getvalue()

    def get_url(self, params_dict):
        """ Get presigned url. Expires in 300 sec."""
        url = self.client.generate_presigned_url(
                ClientMethod='get_object',
                Params={
                    "Bucket": params_dict.get('Bucket'),
                    "Key": params_dict.get('Key')
                },
                ExpiresIn=300
            )  # generate file url

        return url

    def delete_file(self, params_dict):
        """ Delete file"""
        try:
            status = self.client.delete_object(
                        Bucket=params_dict.get('Bucket'),
                        Key=params_dict.get('Key')
            )
            params_dict['Bucket'] = ''
            params_dict['Key'] = ''
            return True
        except Exception as e:
            print(e, 'while trying to delete', params_dict)
            return False

    def create_bucket(self, bucket_name):
        """
        Create a bucket. Handle exception if bucket already exists
        """
        try:
            boto_session = boto3.session.Session()
            self.client.create_bucket(
                Bucket=bucket_name,
                CreateBucketConfiguration={
                    'LocationConstraint': boto_session.region_name
                }
            )
        except Exception as e:
            print(e)

        s3_bucket_exists_waiter = self.client.get_waiter('bucket_exists')

        try:
            s3_bucket_exists_waiter.wait(Bucket='mybucket')
            return True
        except Exception as e:
            print(e)
            return False


    def put_object(self, key):
        """ Put a key into a Bucket. """
        try:
            self.client.put_object(Bucket=self.bucket, Key=key,)
        except Exception as e:
            return False

        return key



import requests
import os, sys

if sys.version_info.major < 3:
    from urllib import url2pathname
else:
    from urllib.request import url2pathname

class LocalFileAdapter(requests.adapters.BaseAdapter):
    """Protocol Adapter to allow Requests to GET file:// URLs

    @todo: Properly handle non-empty hostname portions.
    """

    @staticmethod
    def _chkpath(method, path):
        """Return an HTTP status for the given filesystem path."""
        if method.lower() in ('put', 'delete'):
            return 501, "Not Implemented"  # TODO
        elif method.lower() not in ('get', 'head'):
            return 405, "Method Not Allowed"
        elif os.path.isdir(path):
            return 400, "Path Not A File"
        elif not os.path.isfile(path):
            return 404, "File Not Found"
        elif not os.access(path, os.R_OK):
            return 403, "Access Denied"
        else:
            return 200, "OK"

    def send(self, req, **kwargs):  # pylint: disable=unused-argument
        """Return the file specified by the given request

        @type req: C{PreparedRequest}
        @todo: Should I bother filling `response.headers` and processing
               If-Modified-Since and friends using `os.stat`?
        """
        path = os.path.normcase(os.path.normpath(url2pathname(req.path_url)))
        response = requests.Response()

        response.status_code, response.reason = self._chkpath(req.method, path)
        if response.status_code == 200 and req.method.lower() != 'head':
            try:
                response.raw = open(path, 'rb')
            except (OSError, IOError) as err:
                response.status_code = 500
                response.reason = str(err)

        if isinstance(req.url, bytes):
            response.url = req.url.decode('utf-8')
        else:
            response.url = req.url

        response.request = req
        response.connection = self

        return response

    def close(self):
        pass



import time
from random import random
from functools import wraps

try:
    from django.db import transaction
except Exception as e:
    print(e)


def wait_and_process_transaction(max_tries=5):
    """ Function decorator That applies a standard workflow to wait and execute sql statement in a transaction
    - Note this functionality assumes django is installed.
    """
    def wait_and_process_transaction_decorator(method):

        @wraps(method)
        def processor(*args, **kwargs):
            # This should be assigned to different variable to avoid scope issues
            total_tries = max_tries if max_tries > 1 else 1
            num_tries = 0
            updated = False
            while num_tries < total_tries:
                try:
                    with transaction.atomic():
                        result = method(*args, **kwargs)
                        return result
                    # NOTE Important to break of transaction id succwssful
                    updated = True
                    break
                except Exception as e:
                    num_tries = num_tries + 1
                    print("Trouble executing '%s', Num Tries: %d."%(method.__name__, num_tries), e)
                    time.sleep(random())

            if not updated:
                # raise exception
                print("Failed. Unable to execute '%s', Aborting after %d tries"%(method.__name__, max_tries))
                return None
        return processor
    return wait_and_process_transaction_decorator



import copy, logging, unicodedata, yaml, os
import re
from dateutil import tz, parser

logger = logging.getLogger('datefinder')


class DateFinder(object):
    """
    Locates dates in a text
    """

    def __init__(self, base_date=None):
        self.base_date = base_date

        dirname = os.path.dirname(os.path.dirname(__file__))
        filename = os.path.join(dirname, 'config/datefinder.yaml')
        with open(filename) as file:
            self.settings = yaml.load(file)
            file.close()

        self.DIGITS_MODIFIER_PATTERN = self.settings['DIGITS_MODIFIER_PATTERN']
        self.DIGITS_PATTERN = self.settings['DIGITS_PATTERN']
        self.DAYS_PATTERN = self.settings['DAYS_PATTERN']
        self.MONTHS_PATTERN = self.settings['MONTHS_PATTERN']
        self.TIMEZONES_PATTERN = self.settings['TIMEZONES_PATTERN']
        self.NA_TIMEZONES_PATTERN = self.settings['NA_TIMEZONES_PATTERN']
        self.DELIMITERS_PATTERN = self.settings['DELIMITERS_PATTERN']
        self.TIME_PERIOD_PATTERN = self.settings['TIME_PERIOD_PATTERN']
        self.EXTRA_TOKENS_PATTERN = self.settings['EXTRA_TOKENS_PATTERN']
        self.RELATIVE_PATTERN = self.settings['RELATIVE_PATTERN']
        self.TIME_SHORTHAND_PATTERN = self.settings['TIME_SHORTHAND_PATTERN']
        self.UNIT_PATTERN = self.settings['UNIT_PATTERN']
        self.STRIP_CHARS = self.settings['STRIP_CHARS']
        self.REPLACEMENTS = self.settings['REPLACEMENTS']
        self.TIMEZONE_REPLACEMENTS = self.settings['TIMEZONE_REPLACEMENTS']
        self.TRANSLATIONS = self.settings['TRANSLATIONS']

        self.ALL_TIMEZONES_PATTERN = self.TIMEZONES_PATTERN + '|' + self.NA_TIMEZONES_PATTERN

        ## Time pattern is used independently, so specified here.
        self.TIME_PATTERN = """
        (?P<time>
            ## Captures in format XX:YY(:ZZ) (PM) (EST)
            (
                (?P<hours>\d{{1,2}})
                \:
                (?P<minutes>\d{{1,2}})
                (\:(?P<seconds>\d{{1,2}}))?
                ([\.\,](?P<microseconds>\d{{1,6}}))?
                \s*
                (?P<time_periods>{time_periods})?
                \s*
                (?P<timezones>{timezones})?
            )
            |
            ## Captures in format 11 AM (EST)
            ## Note with single digit capture requires time period
            (
                ## regex in lambda bug: Note re module does not allow repeated group names but regex module does
                (?P<hours2>\d{{1,2}})
                \s*
                (?P<time_periods2>{time_periods})
                \s*
                (?P<timezones2>{timezones})*
            )
        )
        """
        self.TIME_PATTERN = self.TIME_PATTERN.format(
            time_periods=self.TIME_PERIOD_PATTERN,
            timezones=self.ALL_TIMEZONES_PATTERN
        )
        self.TIME_REGEX = re.compile(self.TIME_PATTERN, re.IGNORECASE | re.MULTILINE | re.UNICODE | re.DOTALL | re.VERBOSE)

        self.DATES_PATTERN = """
        (
            (
                {time}
                |
                ## Grab any digits
                (?P<digits_modifier>{digits_modifier})
                |
                (?P<digits>{digits})
                |
                (?P<days>{days})
                |
                (?P<months>{months})
                |
                ## Delimiters, ie Tuesday[,] July 18 or 6[/]17[/]2008
                ## as well as whitespace
                (?P<delimiters>{delimiters})
                |
                ## These tokens could be in phrases that dateutil does not yet recognize
                ## Some are US Centric
                (?P<extra_tokens>{extra_tokens})
            ## We need at least three items to match for minimal datetime parsing
            ## ie 10pm
            ){{3,}}
        )
        """
        self.DATES_PATTERN = self.DATES_PATTERN.format(
            time=self.TIME_PATTERN,
            digits=self.DIGITS_PATTERN,
            digits_modifier=self.DIGITS_MODIFIER_PATTERN,
            days=self.DAYS_PATTERN,
            months=self.MONTHS_PATTERN,
            delimiters=self.DELIMITERS_PATTERN,
            extra_tokens=self.EXTRA_TOKENS_PATTERN
        )
        self.DATE_REGEX = re.compile(self.DATES_PATTERN, re.IGNORECASE | re.MULTILINE | re.UNICODE | re.DOTALL | re.VERBOSE)

    def find_dates(self, text, source=False, index=False, strict=False):

        for date_string, indices, captures in self.extract_date_strings(text, strict=strict):
            as_dt = self.parse_date_string(date_string, captures)
            if as_dt is None:
                ## Dateutil couldn't make heads or tails of it
                ## move on to next
                continue

            returnables = (as_dt,)
            if source:
                returnables = returnables + (date_string,)
            if index:
                returnables = returnables + (indices,)

            if len(returnables) == 1:
                returnables = returnables[0]
            yield returnables

    def _find_and_replace(self, date_string, captures):
        """
        :warning: when multiple tz matches exist the last sorted capture will trump
        :param date_string:
        :return: date_string, tz_string
        """
        # add timezones to replace
        cloned_replacements = copy.copy(self.REPLACEMENTS)  # don't mutate
        # for tz_string in captures.get('timezones', []):  # regex in lambda bug
        #    cloned_replacements.append({tz_string: ' '})  # regex in lambda bug
        tz_string = captures.get('timezones')  # regex in lambda bug
        if tz_string is not None:  # regex in lambda bug
            cloned_replacements.append({tz_string: ' '})  # regex in lambda bug

        date_string = date_string.lower()
        for item in cloned_replacements:
            key = list(item.keys())[0]
            replacement = item[key]
            # we really want to match all permutations of the key surrounded by whitespace chars except one
            # for example: consider the key = 'to'
            # 1. match 'to '
            # 2. match ' to'
            # 3. match ' to '
            # but never match r'(\s|)to(\s|)' which would make 'october' > 'ocber'
            date_string = re.sub(r'(^|\s)' + key + '(\s|$)', replacement, date_string, flags=re.IGNORECASE)

        # Translations support - Change month names to standard english.
        for item in self.TRANSLATIONS:
            key = list(item.keys())[0]
            replacement = item[key]
            date_string = re.sub(key, replacement, date_string, flags=re.IGNORECASE)

        # return date_string, self._pop_tz_string(sorted(captures.get('timezones', [])))  # regex in lambda bug, also Sorting is not needed
        return date_string, self._pop_tz_string([captures.get('timezones', '')])  # regex in lambda bug

    def _pop_tz_string(self, list_of_timezones):
        try:
            tz_string = list_of_timezones.pop()
            # make sure it's not a timezone we
            # want replaced with better abbreviation
            return self.TIMEZONE_REPLACEMENTS.get(tz_string, tz_string)
        except IndexError:
            return ''
        except Exception as e:  # Exception handling for re.groupdict() instead of regex.capturesdict()
            logger.debug(e)
            return ''

    def _add_tzinfo(self, datetime_obj, tz_string):
        """
        take a naive datetime and add dateutil.tz.tzinfo object
        :param datetime_obj: naive datetime object
        :return: datetime object with tzinfo
        """
        if datetime_obj is None:
            return None

        tzinfo_match = tz.gettz(tz_string)
        return datetime_obj.replace(tzinfo=tzinfo_match)

    def parse_date_string(self, date_string, captures):
        # For well formatted string, we can already let dateutils parse them
        # otherwise self._find_and_replace method might corrupt them
        try:
            as_dt = parser.parse(date_string, default=self.base_date)
        except ValueError:
            # replace tokens that are problematic for dateutil
            date_string, tz_string = self._find_and_replace(date_string, captures)

            ## One last sweep after removing
            date_string = date_string.strip(self.STRIP_CHARS)
            ## Match strings must be at least 3 characters long
            ## < 3 tends to be garbage
            if len(date_string) < 3:
                return None

            try:
                logger.debug('Parsing {0} with dateutil'.format(date_string))
                as_dt = parser.parse(date_string, default=self.base_date)
            except Exception as e:
                logger.debug(e)
                as_dt = None
            if tz_string:
                as_dt = self._add_tzinfo(as_dt, tz_string)
        except Exception as e:  # Catching unknown exceptions
            logger.debug(e)
            return None
        return as_dt

    def extract_date_strings(self, text, strict=False):
        """
        Scans text for possible datetime strings and extracts them
        source: also return the original date string
        index: also return the indices of the date string in the text
        strict: Strict mode will only return dates sourced with day, month, and year
        """
        for match in self.DATE_REGEX.finditer(text):
            match_str = match.group(0)
            indices = match.span(0)

            ## Get individual group matches
            # captures = match.capturesdict()  # regex in lambda bug
            captures = match.groupdict()  # regex in lambda bug
            time = captures.get('time')
            digits = captures.get('digits')
            digits_modifiers = captures.get('digits_modifiers')
            days = captures.get('days')
            months = captures.get('months')
            timezones = captures.get('timezones')
            delimiters = captures.get('delimiters')
            time_periods = captures.get('time_periods')
            extra_tokens = captures.get('extra_tokens')

            if strict:
                complete = False
                ## 12-05-2015
                if len(digits) == 3:
                    complete = True
                ## 19 February 2013 year 09:10
                elif (len(months) == 1) and (len(digits) == 2):
                    complete = True

                if not complete:
                    continue

            ## sanitize date string
            ## replace unhelpful whitespace characters with single whitespace
            match_str = re.sub('[\n\t\s\xa0]+', ' ', match_str)
            match_str = match_str.strip(self.STRIP_CHARS)

            ## Save sanitized source string
            yield match_str, indices, captures


def find_dates(
    text,
    source=False,
    index=False,
    strict=False,
    base_date=None
    ):
    """
    Extract datetime strings from text
    :param text:
        A string that contains one or more natural language or literal
        datetime strings
    :type text: str|unicode
    :param source:
        Return the original string segment
    :type source: boolean
    :param index:
        Return the indices where the datetime string was located in text
    :type index: boolean
    :param strict:
        Only return datetimes with complete date information. For example:
        `July 2016` of `Monday` will not return datetimes.
        `May 16, 2015` will return datetimes.
    :type strict: boolean
    :param base_date:
        Set a default base datetime when parsing incomplete dates
    :type base_date: datetime
    :return: Returns a generator that produces :mod:`datetime.datetime` objects,
        or a tuple with the source text and index, if requested
    """
    date_finder = DateFinder(base_date=base_date)
    return date_finder.find_dates(text, source=source, index=index, strict=strict)



# -*- coding: utf-8 -*-
import os, unittest
from datetime import datetime
from shutil import copyfile

from chappie.utils.date_finder import find_dates


class DateFinderTest(unittest.TestCase):
    def setUp(self):
        """To be executed before each test."""
        self.string_of_dates = [
            (u"# entries are due by enero 4th, 2017 at 8:00pm - ok", 1),
            (u"# 09, 2017 - ok", 9),
            (u"# Ejercicio comprendido entre el 01 de Enero y el 31 de Enero 2017 - ok", 1),
            (u"# Ejercicio comprendido entre el 01 de Enero y el 28 de Febrero 2017 - ok", 2),
            (u"# Ejercicio comprendido entre el 01 de Enero y el 31 de Marzo 2017 - ok", 3),
            (u"# Ejercicio comprendido entre el 01 de Enero y el 30 de Abril 2017 - ok", 4),
            (u"# ACUMULADO AL 30 DE ABRIL 2017 - ok", 4),
            (u"# ACUMULADO AL 31 DE AGOSTO  DE 2017 - ok", 8),
            (u"# ACUMUADO  AL 31 DE MARZO DE 2017 - ok", 3),
            (u"# ACUMULADO AL 31 DE ENERO DE 2017 - ok", 1),
            (u"# ACUMULADO AL 28 FEBRERO 2017 - ok", 2),
            (u"# ACUMULADO AL 31 DE JULIO DE 2017 - ok", 7),
            (u"# ACUMULADO AL 31 DE MAYO DE 2017 - ok", 5),
            (u"# ACUMULADO AL 30 DE SEPTIEMBRE DE 2017 - ok", 9),
            (u"Junio/2017 - ok", 6),
            (u"Abril/2017 - ok", 4),
            (u"Agosto/2017 - ok", 8),
            (u"ACUMULADO AL 31 DE JULIO DE 2017 - ok", 7),
            (u"EJERCICIO DE ENERO A ABRIL DEL 2017 - ok", 4),
            (u"EJERCICIO DE ENERO A DICIEMBRE DEL 2016 - ok", 12),
            (u"EJERCICIO DE ENERO A FEBRERO DEL 2017 - ok", 2),
            (u"EJERCICIO DE ENERO A JUNIO DEL 2017 - ok", 6),
            (u"DE ENERO A SEPTIEMBRE DE 2017 - ok", 9),
            (u"BALANCE_20170623174628 - ok", 6),
            (u"2017/3 - ok", 3),
            (u"BALANCE_20170824144418 - ok", 8),
            (u"""COMPAÑIA MINERA  SALI HOCHSCHILD S.A.
            R U T  N 90.853.000-4 false positive
            EXTRACCION DE COBRE, ARRIENDO DE MINAS		     B A L A N C E       T R I B U T A R I O					I V   NIVEL
            ROMAN DIAZ N 205, DEPTO. 301 false positive
            PROVIDENCIA		ACUMUADO  AL 31 DE MARZO DE 2017 ok
            SANTIAGO""", 3),
            (u"""ADMETRICKS SPA
            76.192.113-4  false positive
            BALANCE GENERAL
            Ejercicio comprendido entre el 01 de Enero y el 31 de Marzo 2017	ok		""", 3)
        ]

    def test_date_search(self):
        """check valid dates are returned for a bunch of test strings"""
        for item in self.string_of_dates:
            matches = find_dates(item[0],source=True, index=True, strict=False, base_date=None)
            for match in matches:
                pass
            # last item
            if match:
                period = match[0]
                self.assertIsInstance(period, datetime)
                self.assertEqual(period.month, item[1])



import random
import string


class RandomValue(object):
    """
    Generate a random value by 
    """
    def __init__(self, *args, **kwargs):
        super(RandomValue, self).__init__(*args, **kwargs)

    
    def string_digits(self, string_length=6):
        """
        Generate a random string of letters and digits
        By default string length is 6, could by customized
        """
        letters_and_digits = string.ascii_letters + string.digits
        return ''.join(random.choice(letters_and_digits) for i in range(string_length))

    def string(self, string_length=10):
        """
        Generate a random string of fixed length
        By default string length is 10, could by customized
        """
        letters = string.ascii_letters
        return ''.join(random.choice(letters) for i in range(string_length))

    def string_lower(self, string_length=8):
        """Generate a random string of fixed length """
        letters= string.ascii_lowercase
        return ''.join(random.sample(letters, string_length))

    def rand_string(self, string_length=5, text='abcdefghi'):
        """
        Generate a random string of specific characters 
        """
        return ''.join((random.choice(your_letters) for i in range(string_length)))

    def rand_password(self, string_length=8):
        """
        Generate a random string of letters, digits and special characters
        """
        password_characters = string.ascii_letters + string.digits + string.punctuation
        return ''.join(random.choice(password_characters) for i in range(string_length))



import re


class TextTools():
    """ """

    def clean_text(self, text):
        """Clean text with only characters and numbers."""
        text = re.sub(r'[^\w]', '', text).lower()
        return text

    def get_subdomain(self, host):
        return host.replace('www.', '').split('.')[0]



from .dict_tools import DictTools
from .file_tools import FileTools
from .text_tools import TextTools
from .date_finder import DateFinder
from .request_tools import RequestTools
from .list_tools import ListTools
from .timer_tools import tick_tock
#from .db_tools import wait_and_process_transaction
from .random_value import RandomValue

from .local_file_adapter import LocalFileAdapter



from unittest import TestCase

from chappie.utils.dict_tools import DictTools


class TestDictTools(TestCase):
    """Test Dict Tools"""

    def setUp(self):
        """Some Initial data setup"""
        self.tmp_dict = {
            'item1': '1',
            'item2': '2',
            'group1': {
                'item1': '11',
                'item2': '12',
            }
        }

        self.dict_to_append = {
            'group2': {
                'item2': '22',
            }
        }

        self.dict_to_update = {
            'group1': {
                'item1': 'one-one',
            }
        }

    def tearDown(self):
        """ """

    def test_dict_tools_update(self):
        """Adding and updating dictionary should work at depth"""

        self.tmp_dict = DictTools.update_dict(self.tmp_dict, self.dict_to_append)
        self.assertEqual(self.tmp_dict['group2']['item2'], '22')

        self.tmp_dict = DictTools.update_dict(self.tmp_dict, self.dict_to_update)
        self.assertEqual(self.tmp_dict['group1']['item1'], 'one-one')



from __future__ import division
from pyparsing import (Literal, CaselessLiteral, Word, Combine, Group, Optional,
                       ZeroOrMore, Forward, nums, alphas, oneOf)
import math
import operator


class NumericParser(object):

    def pushFirst(self, strg, loc, toks):
        """"Push first."""
        self.exprStack.append(toks[0])

    def pushUMinus(self, strg, loc, toks):
        """Puch minus."""
        if toks and toks[0] == '-':
            self.exprStack.append('unary -')

    def __init__(self):
        """
        Initialize Numeric Parser.

        expop   :: '^'
        multop  :: '*' | '/'
        addop   :: '+' | '-'
        integer :: ['+' | '-'] '0'..'9'+
        atom    :: PI | E | real | fn '(' expr ')' | '(' expr ')'
        factor  :: atom [ expop factor ]*
        term    :: factor [ multop factor ]*
        expr    :: term [ addop term ]*
        """
        point = Literal(".")
        e = CaselessLiteral("E")
        fnumber = Combine(Word("+-" + nums, nums) +
                          Optional(point + Optional(Word(nums))) +
                          Optional(e + Word("+-" + nums, nums)))
        ident = Word(alphas, alphas + nums + "_$")
        plus = Literal("+")
        minus = Literal("-")
        mult = Literal("*")
        div = Literal("/")
        lpar = Literal("(").suppress()
        rpar = Literal(")").suppress()
        addop = plus | minus
        multop = mult | div
        expop = Literal("^")
        pi = CaselessLiteral("PI")
        expr = Forward()
        atom = ((Optional(oneOf("- +")) +
                 (ident + lpar + expr + rpar | pi | e | fnumber)
                .setParseAction(self.pushFirst)) | Optional(oneOf("- +")) +
                Group(lpar + expr + rpar)
                ).setParseAction(self.pushUMinus)
        # by defining exponentiation as "atom [ ^ factor ]..." instead of
        # "atom [ ^ atom ]...", we get right-to-left exponents, instead of left-to-right
        # that is, 2^3^2 = 2^(3^2), not (2^3)^2.
        factor = Forward()
        factor << atom + \
            ZeroOrMore((expop + factor).setParseAction(self.pushFirst))
        term = factor + \
            ZeroOrMore((multop + factor).setParseAction(self.pushFirst))
        expr << term + \
            ZeroOrMore((addop + term).setParseAction(self.pushFirst))
        # addop_term = ( addop + term ).setParseAction( self.pushFirst )
        # general_term = term + ZeroOrMore( addop_term ) | OneOrMore( addop_term)
        # expr <<  general_term
        self.bnf = expr
        # map operator symbols to corresponding arithmetic operations
        epsilon = 1e-12
        self.opn = {"+": operator.add,
                    "-": operator.sub,
                    "*": operator.mul,
                    "/": operator.truediv,
                    "^": operator.pow}
        self.fn = {"sin": math.sin,
                   "cos": math.cos,
                   "tan": math.tan,
                   "exp": math.exp,
                   "abs": abs,
                   "trunc": lambda a: int(a),
                   "round": round,
                   "sgn": lambda a: abs(a) > epsilon and cmp(a, 0) or 0}

    def evaluateStack(self, s):
        """Evaluate stack."""
        op = s.pop()
        if op == 'unary -':
            return -self.evaluateStack(s)
        if op in "+-*/^":
            op2 = self.evaluateStack(s)
            op1 = self.evaluateStack(s)
            return self.opn[op](op1, op2)
        elif op == "PI":
            return math.pi  # 3.1415926535
        elif op == "E":
            return math.e  # 2.718281828
        elif op in self.fn:
            return self.fn[op](self.evaluateStack(s))
        elif op[0].isalpha():
            return 0
        else:
            return float(op)

    def eval(self, num_string, parseAll=True):
        """Safe eval math string."""
        try:
            self.exprStack = []
            results = self.bnf.parseString(num_string, parseAll)  # NOQA
            val = self.evaluateStack(self.exprStack[:])
        except Exception as e:
            raise(e)
        return val



from unittest import TestCase
import os

from chappie.utils.file_tools import FileTools


class TestFileTools(TestCase):
    """Test Dict Tools"""

    def setUp(self):
        """Some Initial data setup"""
        path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        self.file_tools = FileTools(path + "/resources/test.json")

    def tearDown(self):
        """ """

    def test_file_tools(self):
        """ """
        filename = self.file_tools.generate_unique_filename()

        self.assertTrue(filename, str)
        self.assertTrue(filename[-4:]=="json")
        self.assertFalse(filename=="test.json")

        data = self.file_tools.process_json_file()
        self.assertIsInstance(data, list)
        self.assertTrue(data[0]["data"]=="test")




import boto3, json, os, requests, uuid, sys
from cerberus import Validator
from urllib.request import urlopen
from urllib.parse import urlparse

from .s3_client import S3Client

# TODO Find a solution for loading M2Crypto in Lambda
# from M2Crypto import X509
# from base64 import b64decode
# from M2Crypto.Err import M2CryptoError

class SNSClient():

    SNS_MESSAGE_TYPE_SUB_NOTIFICATION = "SubscriptionConfirmation"
    SNS_MESSAGE_TYPE_NOTIFICATION = "Notification"
    SNS_MESSAGE_TYPE_UNSUB_NOTIFICATION = "UnsubscribeConfirmation"

    MIN_MESSAGE_SIZE_TO_BUFFER = 48000 # in Bytes. Anything bigger than a small json.

    def __init__(self):
        # NOTE Assumoing endpoint url is in aws conf file
        self.sns_client = boto3.client('sns')
        self.sqs_client = boto3.client('sqs')
        self.app_name = ''
        self.topic = ''
        self.protocol = ''
        # self.subscription = ''
        self.endpoint_url = ''
        self.message_filter = '__all__'

    def initialize(self):
        """
        create queues based on Filters
            - apply subscription and filter policies
            - store list of created queues as safe_queue_url_list
        Remove abandoned queues
            - list all queues starting with 'AppName_'
            - For all existing queues not in 'safe_queue_url _list', delete the queues
            NOTE: AWS specification says that there must be a 60 sec delay before creating a queue with the same name as a deleted Queue

        Create a S3 bucket to handle Buffering of large messages

        Update subscription of MicroSerice to the SNS Topics
        """
        safe_queue_url_list = []
        if type(self.message_filter) == list:
            for event in self.message_filter:
                queue_url, token_queue_url = self.initialize_queue(event)
                safe_queue_url_list.append(queue_url)
                safe_queue_url_list.append(token_queue_url)

        existing_queues = self.sqs_client.list_queues(QueueNamePrefix=self.get_queue_name_prefix())

        if existing_queues.get("QueueUrls") is not None:
            for queue_url in existing_queues["QueueUrls"]:
                if queue_url not in safe_queue_url_list:
                    queue_attributes = self.sqs_client.get_queue_attributes(
                            QueueUrl=queue_url,
                            AttributeNames=['All']
                    )
                    queue_arn = queue_attributes["Attributes"].get('QueueArn')

                    try:
                        self.initialize_subscription(self.topic, 'sqs', queue_arn, '__none__')
                        self.sqs_client.delete_queue(
                            QueueUrl=queue_url
                        )
                    except Exception as e:
                        print(e)
        try:
            S3Client().create_bucket(self.get_queue_buffer_bucket())
        except Exception as e:
            print(e)

        self.initialize_subscription(self.topic, self.protocol, self.endpoint_url, self.message_filter)

    def initialize_queue(self, event):
        """
        Create a queue for each event
            if event is not fifo:
                Queue will subscribe to the SNS topic and filter messages by event Name
            If event is fifo:
                No subsctiption. App logig should handle queue management

        A Token Queue is always created while creating a queue.
            App logic should handle creating/deleting tokens and managing concurrency of processess handling the event

        Ensure to purge queue after initialize
            - NOTE that purge queue might make system unstable for up to 60 seconds (Aws docs: https://docs.aws.amazon.com/cli/latest/reference/sqs/purge-queue.html )
        """
        fifo = event.get('fifo', False)
        if fifo == False:
            queue = self.sqs_client.create_queue(
                    QueueName=self.get_queue_name(event),
                    Attributes={
                        "VisibilityTimeout": str(event.get('visibility_timeout', 30)),
                        "ReceiveMessageWaitTimeSeconds": "20",
                    }
                )
        else:
            queue = self.sqs_client.create_queue(
                    QueueName=self.get_queue_name(event),
                    Attributes={
                        "VisibilityTimeout": str(event.get('visibility_timeout', 30)),
                        "ReceiveMessageWaitTimeSeconds": "20",
                        "FifoQueue": "true",
                        "ContentBasedDeduplication": "true"
                    }
                )

        # Note use default visibility. Time between token receive and delete should be very short
        token_queue = self.sqs_client.create_queue(
                QueueName=self.get_token_queue_name(event),
                Attributes={
                    "FifoQueue": "true",
                    "ContentBasedDeduplication": "true",
                    "ReceiveMessageWaitTimeSeconds": "20",
                }
            )

        queue_url = queue.get('QueueUrl')
        token_queue_url = token_queue.get('QueueUrl')

        # purge queues
        try:
            self.sqs_client.purge_queue(QueueUrl=queue_url)
        except Exception as e:
            print(e)

        try:
            self.sqs_client.purge_queue(QueueUrl=token_queue_url)
        except Exception as e:
            print(e)

        if queue_url is not None:  ## NOTE this may be needed for error handling.
            queue_attributes = self.sqs_client.get_queue_attributes(
                    QueueUrl=queue_url,
                    AttributeNames=['All']
            )
            queue_arn = queue_attributes["Attributes"].get('QueueArn')

            if fifo == False:
                self.initialize_subscription(self.topic, 'sqs', queue_arn, [event])

                policy_json = self.allow_sns_to_write_to_sqs(self.topic, queue_arn)

                response = self.sqs_client.set_queue_attributes(
                    QueueUrl = queue_url,
                    Attributes = {
                        'Policy' : policy_json
                    }
                )

            return queue_url, token_queue_url
        else:
            print("Unable to initialize queue for event:", event)
            return None, None

    def initialize_subscription(self, topic, protocol, endpoint_url, message_filter):
        """
        Find subscription. Create one if not found.
        Update filter policy to listen to all events in the message_filter
        """
        if endpoint_url != '' and topic != '':
            try:
                response = self.sns_client.list_subscriptions_by_topic(TopicArn=topic)
                next_token = response.get("NextToken") # TODO loop with nexttoken

                while True:
                    for subscription in response["Subscriptions"]:
                        if subscription["Endpoint"] == endpoint_url:
                            subscription_arn = subscription["SubscriptionArn"]
                            self.update_filter_policy(subscription_arn, message_filter)
                            # self.update_filter_policy(subscription_arn, [item["name"] for item in message_filter])
                            return subscription_arn
                    if next_token is not None:
                        response = self.sns_client.list_subscriptions_by_topic(TopicArn=topic,
                                                                            NextToken=next_token)
                        next_token = response.get("NextToken") # TODO loop with nexttoken
                    else:
                        break

                # NOTE this happens only if no existing subscriptions are found
                if message_filter != '__none__':
                    # Do not subscribe
                    subscription = self.sns_client.subscribe(
                        TopicArn=topic,
                        Protocol=protocol,
                        Endpoint=endpoint_url,
                        ReturnSubscriptionArn=True
                    )
                    subscription_arn = subscription["SubscriptionArn"]
                    self.update_filter_policy(subscription_arn, message_filter)
                    #  self.update_filter_policy(subscription_arn, [item["name"] for item in message_filter])
                    return subscription_arn

            except Exception as e:
                print(e)
                return None
        else:
            return None

    def update_filter_policy(self, subscription_arn, message_filter):
        """
        Implements event filter for the subscription
        Set raw attribute for subscription (reduce message size)
        """
        if len(subscription_arn.split(':')) < 7:
            raise Exception("Invalid subscription: " + subscription_arn)

        if message_filter == "__none__":
            # unsubscribe
            try:
                self.sns_client.unsubscribe(SubscriptionArn=subscription_arn)
            except Exception as e:
                print(e)
            return
        elif message_filter == "__all__":
            filter_policy = {}
        else:
            filter_policy = {
                "fliegen_workflow_event": [item["name"] for item in message_filter]
            }

        try:
            self.sns_client.set_subscription_attributes(
                SubscriptionArn=subscription_arn,
                AttributeName="FilterPolicy",
                AttributeValue=json.dumps(filter_policy)
            )
            self.sns_client.set_subscription_attributes(
                SubscriptionArn=subscription_arn,
                AttributeName="RawMessageDelivery",
                AttributeValue="true"
            )
        except Exception as e:
            print (e)

    def add_process(self, event):
        """
        Check if number of active process is < concurrency limit
        if yes:
            - increment process counter.
            - return process token
        if No:
            - return None
        """

        token_queue_url_dict = self.sqs_client.get_queue_url(QueueName=self.get_token_queue_name(event))
        token_queue_url = token_queue_url_dict["QueueUrl"]

        token_queue_attributes = self.sqs_client.get_queue_attributes(
                                        QueueUrl=token_queue_url,
                                        AttributeNames=['All']
                                    )
        num_processes = int(token_queue_attributes["Attributes"].get('ApproximateNumberOfMessages')) \
                            + int(token_queue_attributes["Attributes"].get('ApproximateNumberOfMessagesNotVisible'))
         # This is a fifo Queue so the number should be accurate. However there may be some discrepancy due to async access

        if num_processes < event.get('concurrency_limit', 1):
            print("New FliegenToken to start task for '%s'"%(event.get('name')))
            return self.sqs_client.send_message(
                QueueUrl=token_queue_url,
                MessageBody=str(uuid.uuid4()),
                MessageGroupId='FliegenToken'
            )
        else:
            return None

    def remove_process(self, event):
        """
        delete token and decrement the active process count
        return None
        """
        token_queue_url_dict = self.sqs_client.get_queue_url(QueueName=self.get_token_queue_name(event))
        token_queue_url = token_queue_url_dict["QueueUrl"]

        response = self.sqs_client.receive_message(
                QueueUrl=token_queue_url,
        )

        # NOTE handle situation when message No message is recieved
        if response.get("Messages") is not None:
            messages_list = response.get("Messages")
            message_handle = messages_list[0]["ReceiptHandle"]

            print("Destroying FliegenToken after processing '%s'"%(event.get('name')))
            return self.sqs_client.delete_message(
                QueueUrl=token_queue_url,
                ReceiptHandle=message_handle
            )
        else:
            return None

    def publish_message(self, message, fliegen_workflow_event='default'):
        """
        Send small messages are sent to the SNS Topic.
        Large messages are stored in a special s3 bucket (with 4 hour expiry) using File manager
            - FileManager returns a params dict with storage data
        The FileManager return value is packaged in the message payload.
        """
        payload = json.dumps(message.get("payload"))
        if sys.getsizeof(payload) > self.MIN_MESSAGE_SIZE_TO_BUFFER:
            file_manager = S3Client()
            file_manager.bucket = self.get_queue_buffer_bucket()
            file_manager.expires = 4  # number of hours to live

            payload = file_manager.upload_file(
                    file_data=payload.encode(),
                    path=self.app_name + '/' + fliegen_workflow_event + '/',
                    filename=str(uuid.uuid4()),
                    params_dict = {
                        "Service": "s3"
                    })
            message["payload"] = payload

        message = {"default": json.dumps(message)} # TODO check if multiple json dumps is needed when using raw_data
        message_attributes = {
            'fliegen_workflow_event': {
                    'StringValue': fliegen_workflow_event,
                    'DataType': 'String'
            }
        }

        try:
            return_value = self.sns_client.publish(
                            TopicArn=self.topic,
                            Message=json.dumps(message),
                            MessageStructure='json',
                            MessageAttributes=message_attributes)
            return return_value
        except Exception as e:
            print(e)
            return None

    def extract_message(self, request):
        """
        NOTE Extract message does not pull actual message if it is buffered in S3
            - this is done for code optimization
        Implements SNS Confirmation workflow.
        Return message data that is packaged in a wrapper (for http endpoint, it's request object)
        """
        try:
            ### TODO Find a solution for loading M2Crypto in Lambda
            ### NOTE This issue is addressed by forcing auth to be passed inside notification_message["Message"]
            ### if self.verify_sns_notification(request):

            notification_message = json.loads(request.body)

            if request.META.get("HTTP_X_AMZ_SNS_MESSAGE_TYPE", None) == self.SNS_MESSAGE_TYPE_SUB_NOTIFICATION or \
                        request.META.get("HTTP_X_AMZ_SNS_MESSAGE_TYPE", None) == self.SNS_MESSAGE_TYPE_UNSUB_NOTIFICATION:
                SubscribeURL = notification_message.get('SubscribeURL')

                if SubscribeURL:
                    confirmation_response = requests.get(SubscribeURL)
                return None
            else:
                if type(notification_message) == str:
                    message = json.loads(notification_message)
                elif type(notification_message) == dict:
                    message = notification_message
                else:
                    print("Unable to interpret Message: ", notification_message)
                    message = None

                return message

        except Exception as e:
            print(e)
            raise(e)

    def retrieve_message(self, event):
        """
        Look into message queue and pull out the next message to process.
        If payload is a FileManager Params dict
            - read actual message form S3 bucket.
            - replace payload in message
        return Message and message Handle
        If no more messages Exist return None
            - NOTE important as this is the termination criteria for recursion
        """
        buffered_message_template = {
            "Service": {"type": "string", "required": True},  # TODO change to report_name
            "Bucket": {"type": "string", "required": True},
            "Key": {"type": "string", "required": True},
        }
        valid_buffered_message = Validator(buffered_message_template, allow_unknown=False)

        queue_url_dict = self.sqs_client.get_queue_url(QueueName=self.get_queue_name(event))
        queue_url = queue_url_dict["QueueUrl"]

        response = self.sqs_client.receive_message(
                QueueUrl=queue_url,
        )

        if response.get("Messages") is not None: #TODO check if this is the right condition
            messages_list = response.get("Messages")

            body = json.loads(messages_list[0]["Body"])
            if type(body) == str:
                body = json.loads(body)
            if body.get("Message") is not None:
                message = json.loads(body.get("Message"))
            else:
                message = body

            payload = message.get('payload')
            if valid_buffered_message.validate(payload):
                file_manager = S3Client()
                file_manager.bucket = self.get_queue_buffer_bucket()

                payload = file_manager.download_file(payload)
                payload = json.loads(payload)
                message["payload"] = payload

            return {
                "message": message,
                "message_handle": messages_list[0]["ReceiptHandle"]
            }
        else:
            print("Processed all messages from queue for event '%s'!"%(event['name']), response)
            return None

    def delete_message(self, event, message_handle):
        """
        NOTE: Buffered data will auto expire in 4 Hrs
        Delete message from queue given the message_handle
        return None
        """
        queue_url_dict = self.sqs_client.get_queue_url(QueueName=self.get_queue_name(event))
        queue_url = queue_url_dict["QueueUrl"]

        return self.sqs_client.delete_message(
            QueueUrl=queue_url,
            ReceiptHandle=message_handle
        )

    def push_message(self, message, queue_url):
        """
        Small messages are pushed into the queue as-is.
        Large messages are stored in a special s3 bucket (with 4 hour expiry) using File manager
            - FileManager returns a params dict with storage data
        The FileManager return value is packaged in the message payload.
        """
        payload = json.dumps(message.get("payload"))
        if sys.getsizeof(payload) > self.MIN_MESSAGE_SIZE_TO_BUFFER:
            file_manager = S3Client()
            file_manager.bucket = self.get_queue_buffer_bucket()
            file_manager.expires = 4  # number of hours to live

            payload = file_manager.upload_file(
                    file_data=payload.encode(),
                    path=self.app_name + '/' + queue_url.split('/')[-1] + '/',
                    filename=str(uuid.uuid4()),
                    params_dict = {
                        "Service": "s3"
                    })

            message["payload"] = payload

        return self.sqs_client.send_message(
            QueueUrl=queue_url,
            MessageBody=json.dumps(message),
            # MessageBody=message,
            MessageGroupId='FliegenPushFifo'
        )

    def get_queue_buffer_bucket(self):
        """
        Create a bucket for each Notification topic (different for each MicroService).
        Messages recieved in the topic will be stored in a file in this bucket.
        ENsure naming is consistent with S3 naming rules
        """
        # app_name = self.app_name
        topic_name = self.topic.split(':')[-1]

        # bucket_name = app_name + '-' + topic_name
        bucket_name = topic_name
        bucket_name = bucket_name.lower()
        bucket_name = bucket_name.replace('_', '-')
        return bucket_name

    def get_queue_name_prefix(self):
        """
        Standardise prefix (helps with searching for queues)
        """
        app_name = self.app_name
        topic_name = self.topic.split(':')[-1]
        if topic_name == '' or topic_name == None:
            raise Exception('Topic Name is weird for some reason. Ouch!, ' + topic_name )

        return app_name + '_' + topic_name + '_'

    def get_queue_name(self, event):
        """
        ensure fifo queues have .fifo suffix
        """
        if event.get('fifo', False):
            return self.get_queue_name_prefix() + event['name'] + '.fifo'
        else:
            return self.get_queue_name_prefix() + event['name'] # + '.fifo'

    def get_token_queue_name(self, event):
        """
        Keep token queue name similar to the queue name
        """
        return self.get_queue_name_prefix() + event['name'] + '_' + 'tokens.fifo'

    def get_event_queue(self, event):
        """
        Get the queue endpoint for a specific event
        """
        queue_url_dict = self.sqs_client.get_queue_url(QueueName=self.get_queue_name(event))
        return queue_url_dict["QueueUrl"]

    def process_event(self, event, message):
        """
        If event is declared as a FIFO queue, the message is pushed into its FIFO queue
            Standard queue messages are auto pushed to SQS via subscription.
        """
        if event.get('fifo', False):
            queue_url = self.get_event_queue(event=event)
            self.push_message(message=message, queue_url=queue_url)

    def allow_sns_to_write_to_sqs(self, topicarn, queuearn):
        """
        Standard policy to allow SNS topic to send message to standard SQS Queue
        """
        policy_document = """{{
                "Version":"2012-10-17",
                "Statement":[
                    {{
                        "Sid":"MyPolicy",
                        "Effect":"Allow",
                        "Principal" : {{"AWS" : "*"}},
                        "Action":"SQS:SendMessage",
                        "Resource": "{}",
                        "Condition":{{
                            "ArnEquals":{{
                                "aws:SourceArn": "{}"
                            }}
                        }}
                    }}
                ]
            }}""".format(queuearn, topicarn)

        return policy_document

    # NOTE These are utils specific to SNS
    def canonical_message_builder(self, content, format):
        """ Builds the canonical message to be verified.

            Sorts the fields as a requirement from AWS

            Args:
                content (dict): Parsed body of the response
                format (list): List of the fields that need to go into the message
            Returns (str):
                canonical message
        """
        m = ""

        for field in sorted(format):
            try:
                m += field + "\n" + content[field] + "\n"
            except KeyError:
                # Build with what you have
                pass

        return str(m)

    def verify_sns_notification(self, request):
        """ Takes a notification request from Amazon push service SNS and verifies the origin of the notification.

            Kudos to Artur Rodrigues for suggesting M2Crypto: http://goo.gl/KAgPPc

            Args:
                request (HTTPRequest): The request object that is passed to the view function
            Returns (bool):
                True if he message passes the verification, False otherwise
            Raises:
                ValueError: If the body of the response couldn't be parsed
                M2CryptoError: If an error raises during the verification process
                URLError: If the SigningCertURL couldn't be opened
        """
        cert = None
        pubkey = None
        canonical_message = None
        canonical_sub_unsub_format = ["Message", "MessageId", "SubscribeURL", "Timestamp", "Token", "TopicArn", "Type"]
        canonical_notification_format = ["Message", "MessageId", "Subject", "Timestamp", "TopicArn", "Type"]
        content = json.loads(request.body)

        if content.get("Signature"):
            decoded_signature = b64decode(content.get("Signature"))

            # Depending on the message type, canonical message format varies: http://goo.gl/oSrJl8
            if request.META.get("HTTP_X_AMZ_SNS_MESSAGE_TYPE", None) == self.SNS_MESSAGE_TYPE_SUB_NOTIFICATION or \
                    request.META.get("HTTP_X_AMZ_SNS_MESSAGE_TYPE", None) == self.SNS_MESSAGE_TYPE_UNSUB_NOTIFICATION:

                canonical_message = self.canonical_message_builder(content, canonical_sub_unsub_format)

            elif request.META.get("HTTP_X_AMZ_SNS_MESSAGE_TYPE", None) == self.SNS_MESSAGE_TYPE_NOTIFICATION:

                canonical_message = self.canonical_message_builder(content, canonical_notification_format)

            else:
                raise ValueError("Message Type (%s) is not recognized" % request.META.get("HTTP_X_AMZ_SNS_MESSAGE_TYPE", None))

            # Load the certificate and extract the public key
            cert = X509.load_cert_string(str(urlopen(content["SigningCertURL"]).read()))
            pubkey = cert.get_pubkey()
            pubkey.reset_context(md='sha1')
            pubkey.verify_init()

            # Feed the canonical message to sign it with the public key from the certificate
            pubkey.verify_update(canonical_message)

            # M2Crypto uses EVP_VerifyFinal() from openssl as the underlying verification function.
            # http://goo.gl/Bk2G36: "EVP_VerifyFinal() returns 1 for a correct signature, 0 for failure and -1
            # if some other error occurred."
            verification_result = pubkey.verify_final(decoded_signature)

            if verification_result == 1:
                return True
            elif verification_result == 0:
                return False
            else:
                raise M2CryptoError("Some error occured while verifying the signature.")
        else:
            return True # For localstack testing



""" Rollbar Notifier """
import sys
import os


import rollbar


class RollbarNotifier():
    """
        from chappie.utils import RollbarNotifier
    """

    def publish_exception(cls, exception_info, e, level='info', extra_data={}):    # pragma: no cover
        """ Publish exception details to rollbar """
        (exc_type, exc_obj, exc_tb) = exception_info

        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        message = {
            'Error': str(exc_type),
            'Source File': str(fname),
            'Line Number': str(exc_tb.tb_lineno),
            'Details': str(e)
        }

        for index, value in extra_data.items():
            message[index] = str(value)

        if os.environ.get('ENV') != 'local':
            rollbar.report_exc_info(extra_data=extra_data)
        else:
            print(message)

        return True


    def publish_message(cls, message, level='info', extra_data={}):
        """ Send a simple message to rollbar """
        message = {
            'message': message
        }
        message.update(extra_data)
        if os.environ.get('ENV') != 'local':
            rollbar.report_message('Got an IOError in the main loop', 'warning')
        else:
            print(message)

        return True



# -*- coding: utf-8 -*-
import os, unittest
from datetime import datetime
from shutil import copyfile

from chappie.utils.date_finder import find_dates


class TestFileManager(unittest.TestCase):
    def setUp(self):
        """ """



import time
from functools import wraps

def tick_tock(debug=False):
    """ Function decorator to compute runtime of a particular method.
        use @tick_tock in front of a function definition
        tick_tock should include a boolean argument 'debug' to suppress or display runtime data
    """
    def tick_tock_decorator(method):

        @wraps(method)
        def timed(*args, **kwargs):
            if debug == True:
                ts = time.time()
                result = method(*args, **kwargs)
                te = time.time()

                print ('%r.%r --- %2.6f seconds ---' % \
                    (method.__module__, method.__name__, te-ts))
                return result
            else:
                result = method(*args, **kwargs)
                return result
        return timed
    return tick_tock_decorator



import json, os, uuid, re
from .text_tools import TextTools

class FileTools():
    """ """

    def __init__(self, filename):
        """Initialize FileTools."""
        self.filename = filename

    def generate_unique_filename(self):
        """Generate unique filename."""
        unique_id = uuid.uuid4().hex
        filename = TextTools().clean_text(os.path.splitext(self.filename)[0])
        extension = os.path.splitext(self.filename)[1][1:]
        new_filename = "%s-%s.%s" % (filename, unique_id, extension)
        return new_filename

    def process_json_file(self):
        """Return processed json file dict."""
        with open(self.filename) as file:
            json_file_bytes = file.read()
            file.close()

        return json.loads(json_file_bytes)



# -*- coding: utf-8 -*-
import os, unittest
from chappie.client.local import LocalClient

class TestLocalClient(unittest.TestCase):
    def setUp(self):
        """ """
        self.client = LocalClient()
        self.client.folder = "/tmp/"
        self.params_dict = {
            "Service": 'local'
        }
        self.params_dict = self.client.upload_file(b'test', 'chappie-test/', 'test.txt', self.params_dict)

    def test_file_upload(self):
        """Test upload of file data"""

        self.assertIsInstance(self.params_dict, dict)
        self.assertEqual(self.params_dict.get('Filename'), '/tmp/chappie-test/test.txt')
        self.assertEqual(self.params_dict.get('Service'), 'local')

    def test_file_download(self):
        """Test download of file data"""
        # params_dict = self.client.upload_file(b'test', '/tmp/chappie-test/', 'test.txt', self.params_dict)

        filedata = self.client.download_file(self.params_dict)
        self.assertIsInstance(filedata, bytes)
        self.assertEqual(filedata.decode(), 'test')

        self.params_dict["Filename"] = "non_existent_file.txt"
        filedata = self.client.download_file(self.params_dict)
        self.assertIsInstance(filedata, bytes)
        self.assertEqual(filedata.decode(), '')

    def test_get_url(self):
        """Test get file access url"""
        # params_dict = self.client.upload_file(b'test', '/tmp/chappie-test/', 'test.txt', self.params_dict)

        url = self.client.get_url(self.params_dict)
        self.assertGreater(len(url), 0)

        self.params_dict = {}
        url = self.client.get_url(self.params_dict)
        self.assertEqual(len(url), 0)

    def test_delete_file(self):
        """Test file delete"""
        # params_dict = self.client.upload_file(b'test', '/tmp/chappie-test/', 'test.txt', self.params_dict)

        return_params = self.client.delete_file(self.params_dict)
        self.assertIsInstance(return_params, dict)

        # return false if file is missing - the params needs to be cleaned up
        status = self.client.delete_file(self.params_dict)
        self.assertIsInstance(return_params, dict)

        # return true if Filename is missing - because there is nothing to do
        self.params_dict = {}
        status = self.client.delete_file(self.params_dict)
        self.assertIsInstance(return_params, dict)



import os


from chappie.client.rollbar import RollbarNotifier


class LogNotifier():
    def __init__(self, *args, **kwargs):
        super(LogNotifier, self).__init__(*args, **kwargs)

        self.notifier_service = os.environ.get('CHAPPIE_NOTIFIER_SERVICE', 'rollbar')    

        if self.notifier_service  == "rollbar":
            self.client = RollbarNotifier()
        else:
            self.client = None
            raise ValueError("Notifier Service (%s) is not supported" % (self.notifier_service))

    def publish_exception(self, exception_info, e, level='info', extra_data={}):
        """  call method from the selected client """
        self.client.publish_exception(exception_info, e, level='info', extra_data={})
        return True

    def publish_message(self, message, level='info', extra_data={}):
        """  call method from the selected client """
        self.client.publish_message(message, level='info', extra_data={})
        return True



from unittest import TestCase
import os

from chappie.utils.text_tools import TextTools


class TestFileTools(TestCase):
    """Test Dict Tools"""

    def setUp(self):
        """Some Initial data setup"""
        self.text_tools = TextTools()

    def tearDown(self):
        """ """

    def test_file_tools(self):
        """ """
        text = "Ñjavhb4389343 cjhla92834r*YQ*&Y#R&@ERF4  F73RT8fWx"
        clean_text = self.text_tools.clean_text(text)

        self.assertEqual(clean_text, "ñjavhb4389343cjhla92834ryqyrerf4f73rt8fwx")




import sys
import collections


class DictTools():
    """ Dicttools provides some dictionary helper functions """
    @staticmethod
    def update_dict(dictionary, new_record):
        """Update dict.
        - Updates the dictionary (add new or update data if record already exists)
        """
        # Python3 replace iteritems() with items()
        if sys.version_info < (3, 0):
            key_value_list = new_record.iteritems()
        else:
            key_value_list = new_record.items()

        for k, v in key_value_list:
            if isinstance(v, collections.Mapping):
                r = DictTools.update_dict(dictionary.get(k, {}), v)
                dictionary[k] = r
            else:
                dictionary[k] = new_record[k]
        return dictionary



import os
import pathlib
import errno


class LocalClient():

    def __init__(self):
        """ """
        self.bucket = ""
        self.base_dir = ""

    def upload_file(self, file_data, path, filename, params_dict):
        """ Save Bytes array data into file and return FileManager Params."""
        abs_path = os.path.join(self.base_dir, 'file_manager', self.bucket, path)
        if not os.path.exists(abs_path):
            try:
                os.makedirs(abs_path)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise e

        with open(os.path.join(abs_path, filename), 'wb') as file:
            file.write(file_data)
            file.close()

        # params_dict['Filename'] = os.path.join(self.bucket, path, filename)
        return_params_dict = params_dict.copy()
        return_params_dict['Bucket'] = self.bucket
        return_params_dict['Key'] = '%s/%s' % (path, filename)
        return return_params_dict

    def download_file(self, params_dict):
        """ Return file data as a Bytes array"""
        filename = params_dict.get('Key')
        bucket = params_dict.get('Bucket')
        abs_path = os.path.join(self.base_dir, 'file_manager', bucket, filename)

        file_data = b''
        if filename:
            try:
                with open(abs_path, 'rb') as file:
                    # if file.exists():
                    file_data = file.read()
                    file.close()
            except:
                file_data = b''

        return file_data

    def get_url(self, params_dict):
        """ Return local file as a url resource"""
        filename = params_dict.get('Key')
        bucket = params_dict.get('Bucket')
        abs_path = os.path.join(self.base_dir, 'file_manager', bucket, filename)
        url = ''
        if filename:
            url = pathlib.Path(abs_path).as_uri()

        return url

    def delete_file(self, params_dict):
        """ Delete file """
        filename = params_dict.get('Key')
        bucket = params_dict.get('Bucket')
        abs_path = os.path.join(self.base_dir, 'file_manager', bucket, filename)
        master_bucket = os.path.join(self.base_dir, 'file_manager', self.bucket)
        file_path = os.path.dirname(abs_path)

        if filename is not None:
            try:
                os.remove(abs_path)
                params_dict["Key"] = ""

                # delete subfolders if they are empty
                for root, dirs, files in os.walk(master_bucket, topdown=False):
                    
                    # for macosx
                    try:
                        os.remove(os.path.join(root, '.DS_Store'))
                    except:
                        pass

                    if len(os.listdir(root)) == 0 and root != master_bucket:
                        os.rmdir(root)

                return params_dict
            except OSError:  ## if failed, report it back to the user ##
                return params_dict
        else:
            params_dict["Key"] = ""
            return params_dict

    def create_bucket(self, bucket_name):
        """
        Create a bucket. Handle exception if bucket already exists
        """
        full_path = os.path.join(self.base_dir, 'file_manager', bucket_name)
        try:
            os.makedirs(full_path)
        except OSError as e:
            raise e

        return full_path

    def put_object(self, key):
        """ Put a key into a Bucket. """
        abs_path = os.path.join(self.base_dir, 'file_manager', self.bucket, key)
        if not os.path.exists(abs_path):
            try:
                os.makedirs(abs_path)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise e

        return abs_path



from bisect import bisect_left

class ListTools():
    def index(self, a, x):
        ''' Locate the leftmost value exactly equal to x
            NOTE: It seems important to ensure ListTools().index works.
            There seems to be a difference in the sorting in query vs in Python.
            may be due to locale setting in Mysql??
        '''
        i = bisect_left(a, x)
        if i != len(a) and a[i] == x:
            return i
        raise ValueError

    def find_lt(self, a, x):
        'Find rightmost value less than x'
        i = bisect_left(a, x)
        if i:
            return a[i-1]
        raise ValueError

    def find_le(self, a, x):
        'Find rightmost value less than or equal to x'
        i = bisect_right(a, x)
        if i:
            return a[i-1]
        raise ValueError

    def find_gt(self, a, x):
        'Find leftmost value greater than x'
        i = bisect_right(a, x)
        if i != len(a):
            return a[i]
        raise ValueError

    def find_ge(self, a, x):
        'Find leftmost item greater than or equal to x'
        i = bisect_left(a, x)
        if i != len(a):
            return a[i]
        raise ValueError

    def chunks(self, l, n):
        'generate chunks of max_length n from list l'
        # For item i in a range that is a length of l,
        for i in range(0, len(l), n):
            # Create an index range for l of n items:
            yield l[i:i+n]





class RequestTools():
    """ """

    def get_auth_from_request(self, request):
        """Estract auth_data from request object"""
        try:
            auth_data = request.auth.decode()
        except Exception as e:
            print(e)
            auth_data = ''

        return auth_data
